/*
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */
#include <stdio.h>
#include <unistd.h>
#include <ogcsys.h>
#include <gccore.h>
#include <ogc/cache.h>
#include <string.h>
#include <stdint.h>
#include <ogc/lwp_watchdog.h>
#include <time.h>

#include "dvd.h"
#include "dvdreadpatch.h"
#include "highplugin_bin.h"
#include "lowplugin_bin.h"
#include "debugplugin_bin.h"
#include "gcplugin.h"
#include "debugprintf.h"
#include "loadapp.h"
#include "systemconfig.h"
#include "patchcode.h"
#include "asm.h"
#include "processor.h"
#include "gameloader.h"

#include "Toy.h"

#include "GCcodehandler_bin.h"

extern void __exception_closeall();

void nothing()
{
}

extern u32 hookselect;
extern u32 hook2select;
extern u32 ocarinaselect;
extern u32 debuggerselect;

extern u32 patched_read;
extern bool patched_readMIOS;
extern bool patched_seek;
extern bool patched_readID;
extern bool patched_readIDMIOS;
extern bool patched_readaudio;
extern bool patched_audioconfig;
extern bool patched_audiostatus;


// Only for debugging:
extern bool patched_reporterror;
extern bool patched_reset;
extern bool patched_stop_motor;
extern bool patched_stream_write;

extern u32 streaming;

int patch__GXSetVAT(u8 *buf, u32 size) 
{
	u32 i;
	u32 match_count = 0;
	const u32 old_table[32] = {
	0x8142ce00, 0x39800000, 0x39600000, 0x3ce0cc01,
	0x48000070, 0x5589063e, 0x886a04f2, 0x38000001,
	0x7c004830, 0x7c600039, 0x41820050, 0x39000008,
	0x99078000, 0x61230070, 0x380b001c, 0x98678000,
	0x61250080, 0x388b003c, 0x7cca002e, 0x61230090,
	0x380b005c, 0x90c78000, 0x99078000, 0x98a78000,
	0x7c8a202e, 0x90878000, 0x99078000, 0x98678000,
	0x7c0a002e, 0x90078000, 0x396b0004, 0x398c0001
	};
	const u32 new_table[32] = {
	0x8122ce00, 0x39400000, 0x896904f2, 0x7d284b78,
	0x556007ff, 0x41820050, 0x38e00008, 0x3cc0cc01,
	0x98e68000, 0x61400070, 0x61440080, 0x61430090,
	0x98068000, 0x38000000, 0x80a8001c, 0x90a68000,
	0x98e68000, 0x98868000, 0x8088003c, 0x90868000,
	0x98e68000, 0x98668000, 0x8068005c, 0x90668000,
	0x98068000, 0x556bf87f, 0x394a0001, 0x39080004,
	0x4082ffa0, 0x38000000, 0x980904f2, 0x4e800020
	};
  
	if (size > sizeof(old_table))
	{
		for (i=0; i<size-sizeof(old_table); i+=4) 
		{
			if (!memcmp(buf + i, old_table, sizeof(old_table))) 
			{
				memcpy(buf + i, new_table, sizeof(new_table));
				i += sizeof(new_table);
				match_count++;
			}
		}
	}
	return match_count;
}

//---------------------------------------------------------------------------------
void load_handler()
//---------------------------------------------------------------------------------
{
	memcpy((void*)0x80001800,GCcodehandler_bin,GCcodehandler_bin_size);
	memcpy((void *)0x80001800, (void*)0x80000000, 6);
	
	// The codes are copied earlier
	
	DCFlushRange((void*)0x80001800, 0x1800);
	ICInvalidateRange((void*)0x80001800, 0x1800);
}

void load_app(u8 reloader, u8 high, u8 debug_plugin)
{
	u32 plugin_size = 0;
	const u8 *plugin = NULL;
	void *plugin_base = 0;
	bool highplugin = false;
	int rv;

	void (*app_init)(void (*report)(const char* fmt, ...));
	int (*app_main)(void** dst, int* size, int* offset);
	void* (*app_final)();

	void (*app_entry)(void(**init)(void (*report)(const char* fmt, ...)),
		int (**main)(), void *(**final)());

	static char apploader_header[32] __attribute__((aligned(32)));

	settime(secs_to_ticks(time(NULL) - 946684800));
	
	// Select which plugin to use + memory setup
	if (
		debug_plugin == 1
		|| 
		(
		// Don't use high plugin or its memory setup for retail discs, except it's using the plugin debugger
		backuploading &&
		 (
		// According to emu_kidid Pokemon Box: Ruby and Sapphire writes some data to 0x80001800 and 0x80002600, and reads it. The game is confirmed to not work with the low plugin.
		(memcmp((void *) GAMECODE_ADDRESS, "GPX", 3) == 0 && memcmp((void *)((u32)GAMECODE_ADDRESS+4),"01" , 2) == 0 )
		// Action Replay
		|| 	memcmp((void *) GAMECODE_ADDRESS, "GNHE", 4) == 0
#ifdef DONTTELL
		|| ( backuploading && (hookselect != 0 || hook2select != 0) && (ocarinaselect != 0 || debuggerselect != 0) )
#endif
		|| (high == 1)
		 )
		)
	   )	
	{
		if (debug_plugin == 1)
		{
			plugin = debugplugin_bin;
			plugin_size = debugplugin_bin_size;
			printf("Using debug plugin...\n");
			debug_printf("Using debug plugin...\n");
		} else
		{
			plugin = highplugin_bin;
			plugin_size = highplugin_bin_size;
			printf("Using high plugin...\n");
			debug_printf("Using high plugin...\n");
		}
		plugin_base = HIGH_PLUGIN_BASE;
		setup_sys_config(-1);
		highplugin = true;
	} else {
		plugin_size = lowplugin_bin_size;
		plugin = lowplugin_bin;
		plugin_base = LOW_PLUGIN_BASE;
		setup_sys_config(0);
		debug_printf("Using low plugin...\n");
		highplugin = false;
	}

	if (memcmp((void *) GAMECODE_ADDRESS, "GNHE", 4) == 0 && reloader == 0 && backuploading)
	{
		printf("Action Replay won't work without reloader, overriding setting...\n");
		reloader = 1;
	}

	rv = dvd_read(apploader_header, sizeof(apploader_header), APPLOADER_HEADER_OFFSET);
	DVD_CHECK();

	debug_printf("Reading apploader address 0x%x, len 0%x, offset 0x%x\n",
		(void*)APPLOADER_ADDRESS,
		((*(unsigned long*)(apploader_header + 0x14)) + 31) & ~31,
		APPLOADER_OFFSET);
	rv = dvd_read((void*)APPLOADER_ADDRESS,
		((*(unsigned long*)(apploader_header + 0x14)) + 31) & ~31,
		APPLOADER_OFFSET);
	DVD_CHECK();

	app_entry = (void (*)(void(**)(void (*)(const char*, ...)), int (**)(), void *(**)()))(*(unsigned long*)(apploader_header + 0x10));
	app_entry(&app_init, &app_main, &app_final);
	
	debug_printf("Calling app_init...\n");
	app_init((void (*)(const char*, ...))nothing);

	if (backuploading || debug_plugin == 1)
	{
		/* Install DVD replacement functions. */
		memcpy(plugin_base, plugin, plugin_size);
		debug_printf("Plugin copied...\n");

		u32 var_offset;
		if (highplugin)
		{
			if (debug_plugin == 1)
			{
#ifdef DONTTELL
			var_offset = ((u32) plugin_base) + 2*sizeof(gcplugin_jumptable_debug_t) + 64;	// The hook is located after the jumptable before the other vars
#else
			var_offset = ((u32) plugin_base) + 2*sizeof(gcplugin_jumptable_debug_t);
#endif		
			} else
			{
#ifdef DONTTELL
			var_offset = ((u32) plugin_base) + 2*sizeof(gcplugin_jumptable_high_t) + 64;	// The hook is located after the jumptable before the other vars
#else
			var_offset = ((u32) plugin_base) + 2*sizeof(gcplugin_jumptable_high_t);
#endif		
			}
		} else
		{
			var_offset = ((u32) plugin_base) + 2*sizeof(gcplugin_jumptable_low_t);
		}
		*(u32 *) (var_offset) = base_offset;
		*(u32 *) (var_offset + sizeof(u32)) = second_disc_offset;
		if (reloader > 0)
		{
			if 	(reloader == 2 
			// 007 Agent under Fire doesn't work with the reloader optimisation
			|| (memcmp((void *) GAMECODE_ADDRESS, "GW7", 3) == 0 && memcmp((void *)((u32)GAMECODE_ADDRESS+4),"69" , 2) == 0 )
				)				
			{
				*(u8 *) (var_offset + 2*sizeof(u32)) = 2;			
				// Don't confuse the user with the reloader, it doesn't work on retail debugging
				if (backuploading)
				{
					printf("Reloader set to force...\n");
				}
			} else
			{
				*(u8 *) (var_offset + 2*sizeof(u32)) = 1;			
				if (backuploading)
				{
					printf("Reloader activated...\n");
				}
			}
		} else
		{
			*(u8 *) (var_offset + 2*sizeof(u32)) = 0;	
		}

		// Ikaruga needs a different Audio Status request handling
		if (memcmp((void *) GAMECODE_ADDRESS, "GIK", 3) == 0 && memcmp((void *)((u32)GAMECODE_ADDRESS+4),"70" , 2) == 0 )
		{
			*(u8 *) (var_offset + 2*sizeof(u32) + 1) = 1;	
			printf("Audio Status Request fix activated...\n");
		} else
		{
			*(u8 *) (var_offset + 2*sizeof(u32) + 1) = 0;	
		}

		if (debug_plugin == 1)
		{
			if (backuploading)
			{
				*(u8 *) (var_offset + 2*sizeof(u32) + 2) = 1;	
			} else
			{
				*(u8 *) (var_offset + 2*sizeof(u32) + 2) = 0;	
			}
		}

#ifdef DONTTELL
		if (highplugin)
		{
			if (debug_plugin == 1)
			{
				if ((hook2select != 0) && (ocarinaselect != 0 || debuggerselect != 0))
				{
					*(u8 *) (var_offset + 2*sizeof(u32) + 3) = 1;	
				} else
				{
					*(u8 *) (var_offset + 2*sizeof(u32) + 3) = 0;	
				}
			} else
			{
				if ((hook2select != 0) && (ocarinaselect != 0 || debuggerselect != 0))
				{
					*(u8 *) (var_offset + 2*sizeof(u32) + 2) = 1;	
				} else
				{
					*(u8 *) (var_offset + 2*sizeof(u32) + 2) = 0;	
				}
			}
			copy_2nd_hook((void *)(var_offset - 64));
		}
#endif
		debug_printf("Plugin variables set...\n");
	}

	if (backuploading || debug_plugin == 1)
	{
		DCFlushRange(plugin_base, plugin_size);
		ICInvalidateRange(plugin_base, plugin_size);
	}

	toy_reset();
	u32 GXSetVAT_count = 0;

	bool hookpatched = false;
	while (1)
	{
		void* dst = 0;
		int len = 0,offset = 0;
		debug_printf("Requesting next section(app_main)...\n");
		int res = app_main(&dst, &len, &offset);

		if (!res)
			break;
		debug_printf("Reading section buffer 0x%x, len 0%x, offset 0x%x\n", dst, len, offset);
		rv = dvd_read(dst, len, offset);
		DVD_CHECK();
		
#ifdef DONTTELL
		if ((hookselect != 0) && (ocarinaselect != 0 || debuggerselect != 0))
#else
		if (!backuploading && hookselect != 0 && (ocarinaselect != 0 || debuggerselect != 0))
#endif
		{
			if (dogamehooks(dst,len))
			{
				hookpatched = true;
			}
		}

		dvd_patchread(dst, len, (gcplugin_jumptable_high_t *)plugin_base, 0, (char *) GAMECODE_ADDRESS, backuploading || debug_plugin == 1, highplugin, debug_plugin == 1);

		toy_register_apploader_file_loading(dst, len);
		
		// Patch for Wind Waker freeze on minimap
		if (memcmp((void *) GAMECODE_ADDRESS, "GZL", 3) == 0)
		{
			GXSetVAT_count+=patch__GXSetVAT(dst, len);
		}
		
		debug_printf("Patching section complete...\n");

		DCFlushRange(dst, len);
		ICInvalidateRange(dst, len);
	}
	debug_printf("All sections loaded...\n");

	if (memcmp((void *) GAMECODE_ADDRESS, "GZL", 3) == 0)
	{
		if (GXSetVAT_count > 0)
		{
			printf("__GXSetVAT patched\n");
			debug_printf("__GXSetVAT patched\n");
		} else
		{
			printf("Warning: __GXSetVAT not patched\n");
			debug_printf("Warning: __GXSetVAT not patched\n");
		}
	}
	
	if (backuploading || debug_plugin == 1)
	{
		DCFlushRange(plugin_base, plugin_size);
		ICInvalidateRange(plugin_base, plugin_size);
	}
	
	char text[32];
	if (backuploading)
	{
		snprintf(text, 32, "WARNING: Could");
	} else
	{
		snprintf(text, 32, "WARNING: Would");
	}

	if (patched_read > 1)
	{
		printf("The game uses dvd read function %u\n", patched_read);
	}

	if (patched_read == 0 || !patched_seek || !patched_readID 
	|| !patched_readaudio || !patched_audioconfig || !patched_audiostatus
	|| (debug_plugin == 1 &&
	(!patched_reporterror || !patched_reset || !patched_stop_motor || !patched_stream_write))
	)
	{
		printf("\n");
		if (!patched_read)
		{
			if (patched_readMIOS)
			{
				printf("%s patch read, but the reloader wouldn't\n", text);
			} else
			{
				printf("%s not patch read\n", text);
			}
		}
		
		if (!patched_readID)
		{
			if (patched_readIDMIOS)
			{
				printf("%s patch read id, but the reloader wouldn't\n", text);
			} else
			{
				printf("%s not patch read id\n", text);
			}
		}
		
		if (!patched_seek)
		{
			printf("%s not patch seek\n", text);
		}

		if (streaming != 0)
		{
			if (!patched_readaudio)
			{
				printf("%s not patch read audio\n", text);
			}

			if (!patched_audioconfig)
			{
				printf("%s not patch audio config\n", text);
			}

			if (!patched_audiostatus)
			{
				printf("%s not patch audio status\n", text);
			}
		}

		if (debug_plugin == 1)
		{
			if (!patched_reporterror)
			{
				printf("%s not patch report error\n", text);
			}

			if (!patched_reset)
			{
				printf("%s not patch reset\n", text);
			}

			if (!patched_stop_motor)
			{
				printf("%s not patch stop motor\n", text);
			}

			if (!patched_stream_write)
			{
				printf("%s not patch stream write\n", text);
			}
		}

		printf("\n");

		sleep(15);
	}
	
#ifdef DONTTELL
	if ((hookselect != 0) && (ocarinaselect != 0 || debuggerselect != 0))
#else
	if (!backuploading && hookselect != 0 && (ocarinaselect != 0 || debuggerselect != 0))
#endif
	{
		if (!hookpatched)
		{
			printf("Hook not patched\n");
			sleep(5);
		} else
		{
			//printf("Hook patched\n");
			//sleep(2);
		}
		load_handler();
	}

	debug_printf("Starting game by backuplauncher.\n");

	toy_patch_video_modes();

	sleep(5);
	
	setvideomode();
	
	/* Close Exceptions: */
	unsigned long level = 0;
	_CPU_ISR_Disable(level);
	__exception_closeall();

#ifdef DONTTELL
	if ((hookselect != 0 || hook2select != 0) && (ocarinaselect != 0 || debuggerselect != 0))
#else
	if (!backuploading && hookselect != 0 && (ocarinaselect != 0 || debuggerselect != 0))
#endif
	{		
		appentrypoint = (u32)app_final();
		__asm__(
			"lis %r3, appentrypoint@h\n"
			"ori %r3, %r3, appentrypoint@l\n"
			"lwz %r3, 0(%r3)\n"
			"mtlr %r3\n"
			"lis %r3, 0x8000\n"
			"ori %r3, %r3, 0x18A8\n"
			"mtctr %r3\n"
			"bctr\n"
		);		
	} else
	{	
		void (*entrypoint)() = app_final();
		entrypoint();
	}	
}

